@@ -57,6 +57,11 @@ import traceback as traceback_
 import logging
 import platform
 
+try:
+    from functools import lru_cache
+except ImportError:
+    from backports.functools_lru_cache import lru_cache
+
 import six
 from six.moves import queue
 from six.moves import urllib
@@ -72,13 +77,31 @@ __all__ = ('HTTPRequest', 'HTTPConnection', 'HTTPServer',
            'Gateway', 'get_ssl_adapter_class')
 
 
-if 'win' in sys.platform and hasattr(socket, 'AF_INET6'):
+IS_WINDOWS = platform.system() == 'Windows'
+
+
+if not IS_WINDOWS:
+    import grp
+    import pwd
+    import struct
+
+
+if IS_WINDOWS and hasattr(socket, 'AF_INET6'):
     if not hasattr(socket, 'IPPROTO_IPV6'):
         socket.IPPROTO_IPV6 = 41
     if not hasattr(socket, 'IPV6_V6ONLY'):
         socket.IPV6_V6ONLY = 27
 
 
+if not hasattr(socket, 'SO_PEERCRED'):
+    """
+    NOTE: the value for SO_PEERCRED can be architecture specific, in
+    which case the getsockopt() will hopefully fail. The arch
+    specific value could be derived from platform.processor()
+    """
+    socket.SO_PEERCRED = 17
+
+
 LF = b'\n'
 CRLF = b'\r\n'
 TAB = b'\t'
@@ -122,7 +145,8 @@ class HeaderReader(object):
         Headers which are repeated are folded together using a comma if their
         specification so dictates.
 
-        This function raises ValueError when the read bytes violate the HTTP spec.
+        This function raises ValueError when the read bytes violate the HTTP
+        spec.
         You should probably return "400 Bad Request" if this happens.
         """
         if hdict is None:
@@ -402,7 +426,8 @@ class ChunkedRFile(object):
         self.bytes_read += len(line)
 
         if self.maxlen and self.bytes_read > self.maxlen:
-            raise errors.MaxSizeExceeded('Request Entity Too Large', self.maxlen)
+            raise errors.MaxSizeExceeded(
+                'Request Entity Too Large', self.maxlen)
 
         line = line.strip().split(SEMICOLON, 1)
 
@@ -717,11 +742,13 @@ class HTTPRequest(object):
         self.method = method.upper()
 
         if self.strict_mode and method != self.method:
-            self.simple_response(
-                '400 Bad Request',
-                'Malformed method name: According to RFC 2616 (section 5.1.1) and its successors '
-                'RFC 7230 (section 3.1.1) and RFC 7231 (section 4.1) method names are case-sensitive and uppercase.'
+            resp = (
+                'Malformed method name: According to RFC 2616 '
+                '(section 5.1.1) and its successors '
+                'RFC 7230 (section 3.1.1) and RFC 7231 (section 4.1) '
+                'method names are case-sensitive and uppercase.'
             )
+            self.simple_response('400 Bad Request', resp)
             return False
 
         try:
@@ -760,7 +787,12 @@ class HTTPRequest(object):
             # the validation doesn't take into account, that urllib parses
             # invalid URIs without raising errors
             # https://tools.ietf.org/html/rfc7230#section-5.3.3
-            if _authority != uri or not _port or any((_scheme, _path, _qs, _fragment)):
+            invalid_path = (
+                _authority != uri
+                or not _port
+                or any((_scheme, _path, _qs, _fragment))
+            )
+            if invalid_path:
                 self.simple_response('400 Bad Request',
                                      'Invalid path in Request-URI: request-'
                                      'target must match authority-form.')
@@ -771,21 +803,36 @@ class HTTPRequest(object):
         else:
             uri_is_absolute_form = (scheme or authority)
 
-            if (self.strict_mode and not self.proxy_mode) and uri_is_absolute_form:
-                # https://tools.ietf.org/html/rfc7230#section-5.3.2 (absolute form)
+            disallowed_absolute = (
+                self.strict_mode
+                and not self.proxy_mode
+                and uri_is_absolute_form
+            )
+            if disallowed_absolute:
+                # https://tools.ietf.org/html/rfc7230#section-5.3.2
+                # (absolute form)
                 """Absolute URI is only allowed within proxies."""
-                self.simple_response('400 Bad Request',
-                                     'Absolute URI not allowed if server is not a proxy.')
+                self.simple_response(
+                    '400 Bad Request',
+                    'Absolute URI not allowed if server is not a proxy.',
+                )
                 return False
 
-            if self.strict_mode and not uri.startswith(FORWARD_SLASH) and not uri_is_absolute_form:
-                # https://tools.ietf.org/html/rfc7230#section-5.3.1 (origin_form) and
+            invalid_path = (
+                self.strict_mode
+                and not uri.startswith(FORWARD_SLASH)
+                and not uri_is_absolute_form
+            )
+            if invalid_path:
+                # https://tools.ietf.org/html/rfc7230#section-5.3.1
+                # (origin_form) and
                 """Path should start with a forward slash."""
-                self.simple_response('400 Bad Request',
-                                     'Invalid path in Request-URI: '
-                                     'request-target must contain origin-form '
-                                     'which starts with absolute-path '
-                                     '(URI starting with a slash "/").')
+                resp = (
+                    'Invalid path in Request-URI: request-target must contain '
+                    'origin-form which starts with absolute-path (URI '
+                    'starting with a slash "/").'
+                )
+                self.simple_response('400 Bad Request', resp)
                 return False
 
             if fragment:
@@ -805,7 +852,8 @@ class HTTPRequest(object):
             # But note that "...a URI must be separated into its components
             # before the escaped characters within those components can be
             # safely decoded." http://www.ietf.org/rfc/rfc2396.txt, sec 2.4.2
-            # Therefore, "/this%2Fpath" becomes "/this%2Fpath", not "/this/path".
+            # Therefore, "/this%2Fpath" becomes "/this%2Fpath", not
+            # "/this/path".
             try:
                 # TODO: Figure out whether exception can really happen here.
                 # It looks like it's caught on urlsplit() call above.
@@ -947,10 +995,8 @@ class HTTPRequest(object):
             self.rfile = KnownLengthRFile(self.conn.rfile, cl)
 
         self.server.gateway(self).respond()
+        self.ready and self.ensure_headers_sent()
 
-        if (self.ready and not self.sent_headers):
-            self.sent_headers = True
-            self.send_headers()
         if self.chunked_write:
             self.conn.wfile.write(b'0\r\n\r\n')
 
@@ -991,6 +1037,12 @@ class HTTPRequest(object):
             if ex.args[0] not in errors.socket_errors_to_ignore:
                 raise
 
+    def ensure_headers_sent(self):
+        """Ensure headers are sent to the client if not already sent."""
+        if not self.sent_headers:
+            self.sent_headers = True
+            self.send_headers()
+
     def write(self, chunk):
         """Write unbuffered data to the client."""
         if self.chunked_write and chunk:
@@ -1018,7 +1070,11 @@ class HTTPRequest(object):
             if status < 200 or status in (204, 205, 304):
                 pass
             else:
-                if self.response_protocol == 'HTTP/1.1' and self.method != b'HEAD':
+                needs_chunked = (
+                    self.response_protocol == 'HTTP/1.1'
+                    and self.method != b'HEAD'
+                )
+                if needs_chunked:
                     # Use the chunked transfer-coding
                     self.chunked_write = True
                     self.outheaders.append((b'Transfer-Encoding', b'chunked'))
@@ -1082,13 +1138,16 @@ class HTTPConnection(object):
     rbufsize = io.DEFAULT_BUFFER_SIZE
     wbufsize = io.DEFAULT_BUFFER_SIZE
     RequestHandlerClass = HTTPRequest
+    peercreds_enabled = False
+    peercreds_resolve_enabled = False
 
     def __init__(self, server, sock, makefile=MakeFile):
         """Initialize HTTPConnection instance.
 
         Args:
             server (HTTPServer): web server object receiving this request
-            socket (socket._socketobject): the raw socket object (usually TCP) for this connection
+            socket (socket._socketobject): the raw socket object (usually
+                TCP) for this connection
             makefile (file): a fileobject class for reading from the socket
         """
         self.server = server
@@ -1097,6 +1156,18 @@ class HTTPConnection(object):
         self.wfile = makefile(sock, 'wb', self.wbufsize)
         self.requests_seen = 0
 
+        self.peercreds_enabled = self.server.peercreds_enabled
+        self.peercreds_resolve_enabled = self.server.peercreds_resolve_enabled
+
+        # LRU cached methods:
+        # Ref: https://stackoverflow.com/a/14946506/595220
+        self.resolve_peer_creds = (
+            lru_cache(maxsize=1)(self.resolve_peer_creds)
+        )
+        self.get_peer_creds = (
+            lru_cache(maxsize=1)(self.get_peer_creds)
+        )
+
     def communicate(self):
         """Read each request and respond appropriately."""
         request_seen = False
@@ -1125,52 +1196,28 @@ class HTTPConnection(object):
         except socket.error as ex:
             errnum = ex.args[0]
             # sadly SSL sockets return a different (longer) time out string
-            if (
-                errnum == 'timed out' or
-                errnum == 'The read operation timed out'
-            ):
+            timeout_errs = 'timed out', 'The read operation timed out'
+            if errnum in timeout_errs:
                 # Don't error if we're between requests; only error
                 # if 1) no request has been started at all, or 2) we're
                 # in the middle of a request.
                 # See https://github.com/cherrypy/cherrypy/issues/853
                 if (not request_seen) or (req and req.started_request):
-                    # Don't bother writing the 408 if the response
-                    # has already started being written.
-                    if req and not req.sent_headers:
-                        try:
-                            req.simple_response('408 Request Timeout')
-                        except errors.FatalSSLAlert:
-                            # Close the connection.
-                            return
-                        except errors.NoSSLError:
-                            self._handle_no_ssl(req)
+                    self._conditional_error(req, '408 Request Timeout')
             elif errnum not in errors.socket_errors_to_ignore:
                 self.server.error_log('socket.error %s' % repr(errnum),
                                       level=logging.WARNING, traceback=True)
-                if req and not req.sent_headers:
-                    try:
-                        req.simple_response('500 Internal Server Error')
-                    except errors.FatalSSLAlert:
-                        # Close the connection.
-                        return
-                    except errors.NoSSLError:
-                        self._handle_no_ssl(req)
-            return
+                self._conditional_error(req, '500 Internal Server Error')
         except (KeyboardInterrupt, SystemExit):
             raise
         except errors.FatalSSLAlert:
-            # Close the connection.
-            return
+            pass
         except errors.NoSSLError:
             self._handle_no_ssl(req)
         except Exception as ex:
-            self.server.error_log(repr(ex), level=logging.ERROR, traceback=True)
-            if req and not req.sent_headers:
-                try:
-                    req.simple_response('500 Internal Server Error')
-                except errors.FatalSSLAlert:
-                    # Close the connection.
-                    return
+            self.server.error_log(
+                repr(ex), level=logging.ERROR, traceback=True)
+            self._conditional_error(req, '500 Internal Server Error')
 
     linger = False
 
@@ -1186,6 +1233,22 @@ class HTTPConnection(object):
         req.simple_response('400 Bad Request', msg)
         self.linger = True
 
+    def _conditional_error(self, req, response):
+        """Respond with an error.
+
+        Don't bother writing if a response
+        has already started being written.
+        """
+        if not req or req.sent_headers:
+            return
+
+        try:
+            req.simple_response(response)
+        except errors.FatalSSLAlert:
+            pass
+        except errors.NoSSLError:
+            self._handle_no_ssl(req)
+
     def close(self):
         """Close the socket underlying this connection."""
         self.rfile.close()
@@ -1202,6 +1265,100 @@ class HTTPConnection(object):
             # Apache does, but not today.
             pass
 
+    def get_peer_creds(self):  # LRU cached on per-instance basis, see __init__
+        """Return the PID/UID/GID tuple of the peer socket for UNIX sockets.
+
+        This function uses SO_PEERCRED to query the UNIX PID, UID, GID
+        of the peer, which is only available if the bind address is
+        a UNIX domain socket.
+
+        Raises:
+            NotImplementedError: in case of unsupported socket type
+            RuntimeError: in case of SO_PEERCRED lookup unsupported or disabled
+        """
+        PEERCRED_STRUCT_DEF = '3i'
+
+        if IS_WINDOWS or self.socket.family != socket.AF_UNIX:
+            raise NotImplementedError(
+                'SO_PEERCRED is only supported in Linux kernel and WSL'
+            )
+        elif not self.peercreds_enabled:
+            raise RuntimeError(
+                'Peer creds lookup is disabled within this server'
+            )
+
+        try:
+            peer_creds = self.socket.getsockopt(
+                socket.SOL_SOCKET, socket.SO_PEERCRED,
+                struct.calcsize(PEERCRED_STRUCT_DEF)
+            )
+        except socket.error as socket_err:
+            """Non-Linux kernels don't support SO_PEERCRED.
+
+            Refs:
+            http://welz.org.za/notes/on-peer-cred.html
+            https://github.com/daveti/tcpSockHack
+            msdn.microsoft.com/en-us/commandline/wsl/release_notes#build-15025
+            """
+            six.raise_from(  # 3.6+: raise RuntimeError from socket_err
+                RuntimeError,
+                socket_err,
+            )
+        else:
+            pid, uid, gid = struct.unpack(PEERCRED_STRUCT_DEF, peer_creds)
+            return pid, uid, gid
+
+    @property
+    def peer_pid(self):
+        """Return the id of the connected peer process."""
+        pid, _, _ = self.get_peer_creds()
+        return pid
+
+    @property
+    def peer_uid(self):
+        """Return the user id of the connected peer process."""
+        _, uid, _ = self.get_peer_creds()
+        return uid
+
+    @property
+    def peer_gid(self):
+        """Return the group id of the connected peer process."""
+        _, _, gid = self.get_peer_creds()
+        return gid
+
+    def resolve_peer_creds(self):  # LRU cached on per-instance basis
+        """Return the username and group tuple of the peercreds if available.
+
+        Raises:
+            NotImplementedError: in case of unsupported OS
+            RuntimeError: in case of UID/GID lookup unsupported or disabled
+        """
+        if IS_WINDOWS:
+            raise NotImplementedError(
+                'UID/GID lookup can only be done under UNIX-like OS'
+            )
+        elif not self.peercreds_resolve_enabled:
+            raise RuntimeError(
+                'UID/GID lookup is disabled within this server'
+            )
+
+        user = pwd.getpwuid(self.peer_uid).pw_name  # [0]
+        group = grp.getgrgid(self.peer_gid).gr_name  # [0]
+
+        return user, group
+
+    @property
+    def peer_user(self):
+        """Return the username of the connected peer process."""
+        user, _ = self.resolve_peer_creds()
+        return user
+
+    @property
+    def peer_group(self):
+        """Return the group of the connected peer process."""
+        _, group = self.resolve_peer_creds()
+        return group
+
     def _close_kernel_socket(self):
         """Close kernel socket in outdated Python versions.
 
@@ -1260,7 +1417,9 @@ class HTTPServer(object):
     """The minimum number of worker threads to create (default 10)."""
 
     maxthreads = None
-    """The maximum number of worker threads to create (default -1 = no limit)."""
+    """The maximum number of worker threads to create.
+
+    (default -1 = no limit)"""
 
     server_name = None
     """The name of the server; defaults to ``self.version``."""
@@ -1272,10 +1431,14 @@ class HTTPServer(object):
     features used in the response."""
 
     request_queue_size = 5
-    """The 'backlog' arg to socket.listen(); max queued connections (default 5)."""
+    """The 'backlog' arg to socket.listen(); max queued connections.
+
+    (default 5)."""
 
     shutdown_timeout = 5
-    """The total time, in seconds, to wait for worker threads to cleanly exit."""
+    """The total time to wait for worker threads to cleanly exit.
+
+    Specified in seconds."""
 
     timeout = 10
     """The timeout in seconds for accepted connections (default 10)."""
@@ -1290,7 +1453,7 @@ class HTTPServer(object):
     """
 
     ready = False
-    """An internal flag which marks whether the socket is accepting connections."""
+    """Internal flag which indicating the socket is accepting connections."""
 
     max_request_header_size = 0
     """The maximum size, in bytes, for request headers, or 0 for no limit."""
@@ -1310,7 +1473,17 @@ class HTTPServer(object):
     You must have the corresponding SSL driver library installed.
     """
 
-    def __init__(self, bind_addr, gateway, minthreads=10, maxthreads=-1, server_name=None):
+    peercreds_enabled = False
+    """If True, peer cred lookup can be performed via UNIX domain socket."""
+
+    peercreds_resolve_enabled = False
+    """If True, username/group will be looked up in the OS from peercreds."""
+
+    def __init__(
+        self, bind_addr, gateway,
+        minthreads=10, maxthreads=-1, server_name=None,
+        peercreds_enabled=False, peercreds_resolve_enabled=False,
+    ):
         """Initialize HTTPServer instance.
 
         Args:
@@ -1318,16 +1491,22 @@ class HTTPServer(object):
             gateway (Gateway): gateway for processing HTTP requests
             minthreads (int): minimum number of threads for HTTP thread pool
             maxthreads (int): maximum number of threads for HTTP thread pool
-            server_name (str): web server name to be advertised via Server HTTP header
+            server_name (str): web server name to be advertised via Server
+                HTTP header
         """
         self.bind_addr = bind_addr
         self.gateway = gateway
 
-        self.requests = threadpool.ThreadPool(self, min=minthreads or 1, max=maxthreads)
+        self.requests = threadpool.ThreadPool(
+            self, min=minthreads or 1, max=maxthreads)
 
         if not server_name:
             server_name = self.version
         self.server_name = server_name
+        self.peercreds_enabled = peercreds_enabled
+        self.peercreds_resolve_enabled = (
+            peercreds_resolve_enabled and peercreds_enabled
+        )
         self.clear_stats()
 
     def clear_stats(self):
@@ -1375,10 +1554,27 @@ class HTTPServer(object):
         return '%s.%s(%r)' % (self.__module__, self.__class__.__name__,
                               self.bind_addr)
 
-    def _get_bind_addr(self):
+    @property
+    def bind_addr(self):
+        """Return the interface on which to listen for connections.
+
+        For TCP sockets, a (host, port) tuple. Host values may be any IPv4
+        or IPv6 address, or any valid hostname. The string 'localhost' is a
+        synonym for '127.0.0.1' (or '::1', if your hosts file prefers IPv6).
+        The string '0.0.0.0' is a special IPv4 entry meaning "any active
+        interface" (INADDR_ANY), and '::' is the similar IN6ADDR_ANY for
+        IPv6. The empty string or None are not allowed.
+
+        For UNIX sockets, supply the filename as a string.
+
+        Systemd socket activation is automatic and doesn't require tempering
+        with this variable.
+        """
         return self._bind_addr
 
-    def _set_bind_addr(self, value):
+    @bind_addr.setter
+    def bind_addr(self, value):
+        """Set the interface on which to listen for connections."""
         if isinstance(value, tuple) and value[0] in ('', None):
             # Despite the socket module docs, using '' does not
             # allow AI_PASSIVE to work. Passing None instead
@@ -1394,22 +1590,6 @@ class HTTPServer(object):
                              "Use '0.0.0.0' (IPv4) or '::' (IPv6) instead "
                              'to listen on all active interfaces.')
         self._bind_addr = value
-    bind_addr = property(
-        _get_bind_addr,
-        _set_bind_addr,
-        doc="""The interface on which to listen for connections.
-
-        For TCP sockets, a (host, port) tuple. Host values may be any IPv4
-        or IPv6 address, or any valid hostname. The string 'localhost' is a
-        synonym for '127.0.0.1' (or '::1', if your hosts file prefers IPv6).
-        The string '0.0.0.0' is a special IPv4 entry meaning "any active
-        interface" (INADDR_ANY), and '::' is the similar IN6ADDR_ANY for
-        IPv6. The empty string or None are not allowed.
-
-        For UNIX sockets, supply the filename as a string.
-
-        Systemd socket activation is automatic and doesn't require tempering
-        with this variable""")
 
     def safe_start(self):
         """Run the server forever, and stop it cleanly on exit."""
@@ -1469,12 +1649,14 @@ class HTTPServer(object):
                     host, port, socket.AF_UNSPEC,
                     socket.SOCK_STREAM, 0, socket.AI_PASSIVE)
             except socket.gaierror:
-                if ':' in self.bind_addr[0]:
-                    info = [(socket.AF_INET6, socket.SOCK_STREAM,
-                             0, '', self.bind_addr + (0, 0))]
-                else:
-                    info = [(socket.AF_INET, socket.SOCK_STREAM,
-                             0, '', self.bind_addr)]
+                sock_type = socket.AF_INET
+                bind_addr = self.bind_addr
+
+                if ':' in host:
+                    sock_type = socket.AF_INET6
+                    bind_addr = bind_addr + (0, 0)
+
+                info = [(sock_type, socket.SOCK_STREAM, 0, '', bind_addr)]
 
         if not self.socket:
             msg = 'No socket could be created'
@@ -1537,7 +1719,7 @@ class HTTPServer(object):
         """Create (or recreate) the actual socket object."""
         self.socket = socket.socket(family, type, proto)
         prevent_socket_inheritance(self.socket)
-        if platform.system() != 'Windows':
+        if not IS_WINDOWS:
             # Windows has different semantics for SO_REUSEADDR,
             # so don't set it.
             # https://msdn.microsoft.com/en-us/library/ms740621(v=vs.85).aspx
@@ -1553,7 +1735,12 @@ class HTTPServer(object):
         # If listening on the IPV6 any address ('::' = IN6ADDR_ANY),
         # activate dual-stack. See
         # https://github.com/cherrypy/cherrypy/issues/871.
-        if hasattr(socket, 'AF_INET6') and family == socket.AF_INET6 and host in ('::', '::0', '::0.0.0.0'):
+        listening_ipv6 = (
+            hasattr(socket, 'AF_INET6')
+            and family == socket.AF_INET6
+            and host in ('::', '::0', '::0.0.0.0')
+        )
+        if listening_ipv6:
             try:
                 self.socket.setsockopt(
                     socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
@@ -1563,6 +1750,19 @@ class HTTPServer(object):
                 pass
 
         self.socket.bind(self.bind_addr)
+        # TODO: keep requested bind_addr separate real bound_addr (port is
+        # different in case of ephemeral port 0)
+        self.bind_addr = self.socket.getsockname()
+        if family in (
+            # Windows doesn't have socket.AF_UNIX, so not using it in check
+            socket.AF_INET,
+            socket.AF_INET6,
+        ):
+            """UNIX domain sockets are strings or bytes.
+
+            In case of bytes with a leading null-byte it's an abstract socket.
+            """
+            self.bind_addr = self.bind_addr[:2]
 
     def tick(self):
         """Accept a new connection and put it on the Queue."""
@@ -1656,16 +1856,17 @@ class HTTPServer(object):
                 return
             raise
 
-    def _get_interrupt(self):
+    @property
+    def interrupt(self):
+        """Flag interrupt of the server."""
         return self._interrupt
 
-    def _set_interrupt(self, interrupt):
+    @interrupt.setter
+    def interrupt(self, interrupt):
+        """Perform the shutdown of this server and save the exception."""
         self._interrupt = True
         self.stop()
         self._interrupt = interrupt
-    interrupt = property(_get_interrupt, _set_interrupt,
-                         doc='Set this to an Exception instance to '
-                             'interrupt the server.')
 
     def stop(self):
         """Gracefully shutdown a server that is serving forever."""
@@ -1714,7 +1915,7 @@ class HTTPServer(object):
 
 
 class Gateway(object):
-    """A base class to interface HTTPServer with other systems, such as WSGI."""
+    """Base class to interface HTTPServer with other systems, such as WSGI."""
 
     def __init__(self, req):
         """Initialize Gateway instance with request.
